import time
import warnings
import logging
from pathlib import Path

import fire
import torch
from torch import Tensor

from optimus.models.optimus import OptimusTransformer, OptimusConfig
from optimus.utils import setup_utils


logger = logging.getLogger(__name__)


def sample_top_p(probs: Tensor, top_p: float) -> Tensor:
    """
    Nucleus (top-p) sampling.

    Args:
        probs (Tensor): Output probability distribution of the model. This
            should be the softmax'ed logits.
        top_p (float): Top-p threshold value for sampling.

    Returns:
        Tensor: Sampled output token, as index in the vocabulary.

    """
    probs_sort, probs_idx = torch.sort(probs, dim=-1, descending=True)
    probs_sum = torch.cumsum(probs_sort, dim=-1)
    mask = probs_sum - probs_sort > top_p
    probs_sort[mask] = 0.
    probs_sort.div_(probs_sort.sum(dim=-1, keepdim=True))
    token = torch.multinomial(probs_sort, num_samples=1)
    token = torch.gather(probs_idx, -1, token)
    return token


def main(
    # model + tokenizer
    model_config_path: Path = Path('config.json'),
    weights_path: Path = Path('weights.pth'),
    hf_tokenizer_name: str = 'gpt2',

    # prompt
    prompt: str | None = None,

    # inference args
    max_seq_len: int = 512,
    temperature: float = 0.6,
    top_p: float = 0.9,
    use_fp16: bool = True,
    device: str = 'cuda'
):
    """
    Run the main inference loop for the model.

    Args:
        model_config_path (Path): Path to config for model.
        weights_path (Path): Path to trained model weights.
        hf_tokenizer_name (str): HuggingFace tokenizer. Note: Ensure the same
            tokenizer is used for both training and testing, otherwise the
            results might be undefined.
        prompt (str | None): Prompt to feed to the model. If empty, the user
            will be prompted to enter text on stdin.
        max_seq_len (int): Maximum context length of the model. Make sure the
            value is similar to 'seq_len' used when training, otherwise the
            model will generalize poorly to higher context lengths.
        temperature (float): Temperature value to control randomness in
            sampling.
        top_p (float): Probability threshold used in nucleus sampling.
        use_fp16 (bool): Whether the computation should be in fp16 or
            full-precision fp32. If the model weights are stored as fp16, they
            will be converted to fp32 and vice-versa. If 'use_fp16' is enabled,
            but the model was trained with fp32, performance might drop
            significantly.
        device (str): Device where to run inference. Viable options are 'cpu',
            'cuda', 'cuda:2' etc.

    """

    logger.info(f"Running with:\n"
        f"\t- model config path: '{model_config_path}'\n"
        f"\t- weights file: '{weights_path}'\n"
        f"\t- huggingface tokenizer: '{hf_tokenizer_name}'\n"
        f"\t- prompt: '{prompt[:30] + ('' if len(prompt) <= 30 else '...') if prompt is not None else '(empty)'}'\n"
        f"\t- max context length: {max_seq_len}\n"
        f"\t- temperature: {temperature}\n"
        f"\t- top_p threshold: {top_p}\n"
        f"\t- 16-bit floating-point: {use_fp16}\n"
        f"\t- running inference on device: {device}\n"
        f"Please see '--help' if you want to change these settings")

    # load tokenizer
    tokenizer = setup_utils.create_tokenizer(hf_tokenizer_name)

    # set default tensor type for things like positional encodings
    if use_fp16 is True:
        if device == 'cuda':
            # avoid warnings about set_default_tensor_type() being deprecated
            warnings.simplefilter('ignore')
            torch.set_default_tensor_type(torch.cuda.HalfTensor)
        elif device == 'cpu':
            assert 0 == 1, 'Cannot run 16-bit inference on CPU!'
    else:
        if device == 'cuda':
            warnings.simplefilter('ignore')
            torch.set_default_tensor_type(torch.cuda.FloatTensor)
        elif device == 'cpu':
            warnings.simplefilter('ignore')
            torch.set_default_tensor_type(torch.FloatTensor)

    logger.info('Loading model from disk...')

    # create model and load weights
    config = OptimusConfig.from_json_file(model_config_path)
    model = OptimusTransformer(config)
    model.load_state_dict(torch.load(weights_path, map_location=device), strict=True)

    # alternatively, comment the 3 lines above, and uncomment / change / adapt
    # the following 3 lines to use a HuggingFace model
    # from transformers import AutoTokenizer, AutoModelForCausalLM
    # tokenizer = AutoTokenizer.from_pretrained('gpt2')
    # model = AutoModelForCausalLM.from_pretrained('gpt2')

    model.eval()
    model.to(device)

    logger.info(f'Loaded model on device {device}!')

    _total_params = sum(p.numel() for p in model.parameters())
    logger.info(f'Number of model parameters: {_total_params}')

    # inference loop
    logger.info('Starting inference...')

    if prompt is not None:
        input_sentence = prompt
    else:
        logger.info('Waiting for user input... (prompt to complete)')
        input_sentence = input('User: ')

    # tokenize input
    inp = torch.tensor(tokenizer(input_sentence)['input_ids'], dtype=torch.long)
    inp.unsqueeze_(0) # (1, seq_len)

    seq_len = inp.shape[-1]

    toks_generated = 0
    start_time = time.time()

    with torch.inference_mode():

        while seq_len != max_seq_len:

            # get model output
            output = model(inp)
            output = output.get('logits', output) # (1, seq_len, vocab_sz)
            toks_generated += 1

            # get the logits for the last token
            logits = output[0,-1,:] # (vocab_sz)

            # sample the most relevant token from the probability
            if temperature > 0:
                # if temperature is enabled, nucleus (top-p) sampling
                probs = torch.softmax(logits / temperature, dim=-1)
                token = sample_top_p(probs, top_p)
            else:
                # otherwise, fallback to greedy decoding
                token = torch.argmax(logits, dim=-1)
                token.unsqueeze_(0)

            # check if token is EOS
            if token == tokenizer.eos_token_id:
                break

            # append the token to the input
            inp = torch.cat([inp, token.unsqueeze(0)], dim=-1) # (1, seq_len + 1)

            seq_len = inp.shape[-1]

    print(f"Model output: {''.join(tokenizer.decode(inp.squeeze(0)))}")
    logger.info(f'Tokens / second: {toks_generated / (time.time() - start_time):.2f}')

    logger.info('Finished inference!')


if __name__=='__main__':
    fire.Fire(main)
