import os
import json
import logging
from pathlib import Path
from typing import Optional

from tqdm import tqdm
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from optimus.utils import dist_utils, logging_utils, checkpoint_utils


logger = logging.getLogger(__name__)


class TrainingArguments():
    """
    Training arguments class to hold important switches and knobs related to
    training.

    Args:
        device (torch.device): GPU device on which to train.
        log_steps (int): Log training progress each number of steps. This
            considers number of updates, so gradient_accumulation_steps
            influences this.
        log_on_main_rank_only (bool, defaults to True): Whether the main rank
            should be the only one that logs training information.
        show_progress (bool, defaults to True): Whether to show progress during
            training.
        seed (int): Seed used for reproducibility purposes.
        optimizer (torch.optim.Optimizer): Optimizer used for training.
        lr_scheduler (torch.optim.lr_scheduler.LRScheduler): Scheduler used for
            learning rate adjustment during training.
        num_train_epochs (int): Number of training epochs.
        per_device_batch_size (int): Batch size per device (GPU).
        gradient_accumulation_steps (int): Steps to accumulate the gradient for.
        max_grad_norm (float): Max gradient clipping norm.
        use_fp16 (bool): Whether to train in floating point 16 accuracy. Uses
            bfloat16 if available, otherwise regular fp16.
        checkpoints_dir (Path): Path to where training checkpoints should be
            saved.
        save_steps (int): Save checkpoints every number of steps. If 0, doesn't
            save checkpoints.
        save_limit (int): Limit number of checkpoints to `save_limit`. Starts
            deleting from the oldest when this number is reached.

    """
    def __init__(
        self,
        device: torch.device,
        log_steps: int,
        seed: int,
        optimizer: torch.optim.Optimizer,
        lr_scheduler: torch.optim.lr_scheduler.LRScheduler,
        num_train_epochs: int,
        per_device_batch_size: int,
        gradient_accumulation_steps: int,
        max_grad_norm: float,
        use_fp16: bool,
        checkpoints_dir: Path,
        save_steps: int,
        save_limit: int,
        log_on_main_rank_only: bool = True,
        show_progress: bool = True,
    ):
        self.device = device
        self.log_steps = log_steps
        self.log_on_main_rank_only = log_on_main_rank_only
        self.show_progress = show_progress
        self.seed = seed
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.num_train_epochs = num_train_epochs
        self.per_device_batch_size = per_device_batch_size
        self.gradient_accumulation_steps = gradient_accumulation_steps
        self.max_grad_norm = max_grad_norm
        self.use_fp16 = use_fp16
        self.checkpoints_dir = checkpoints_dir
        self.save_steps = save_steps
        self.save_limit = save_limit


class TrainerState():
    """
    State of a trainer. This contains information that needs to be persisted
    across checkpoints, such as current epoch or current number of steps.
    """
    def __init__(self):
        self.current_epoch = 0

        # this is weight updates, in total, over the whole training loop, with
        # however many epochs; with gradient accumulation steps > 1, there will
        # be more forward + backward passes than weight updates, but this is not
        # captured here; global_steps doesn't care about epochs, but is rather
        # equal to epochs * steps_per_epoch
        self.global_steps = 0

    def to_json_file(self, file: Path):
        """
        Save the current trainer state as a json file.

        Args:
            file (Path): Path where the trainer state json file should be saved.
        """
        print('saving json trainer state')
        with open(file, 'w', encoding='utf-8') as f:
            json.dump(vars(self), f, ensure_ascii=False, indent=4)
        # TODO probably should just be to_json

    @classmethod
    def from_json_file(cls, file: Path | str):
        """
        Reload the state of a trainer from a json file. Note: It is very likely
        that training *will not work as expected* if you resume training from a
        checkpoint with a trainer state that has been manually altered.

        Args:
            file (Path): Path to a json file (usually contained inside a
            checkpoint directory),
        """
        print('loading trainer state from json') #TODO


class Trainer():
    """
    Generic PyTorch trainer implementation.

    Args:
        train_args (TrainingArguments): Training arguments.
        model (nn.Module): Model to train.
        train_dataloader (torch.utils.data.DataLoader): Dataloader used for model
            training.
        eval_dataloader (torch.utils.data.DataLoader): Dataloader used for model
            evaluation.
        tokenizer (PreTrainedTokenizerBase, optional): HuggingFace tokenizer
            used for data collation purposes. Gets saved along with the model.
            If not passed, a default collator is used.

    """

    def __init__(
        self,
        train_args: TrainingArguments,
        model: nn.Module,
        train_dataloader: DataLoader,
        eval_dataloader: DataLoader,
        tokenizer: Optional[PreTrainedTokenizerBase],
    ):
        self.args = train_args
        self.model = model
        self.train_dataloader = train_dataloader
        self.eval_dataloader = eval_dataloader
        self.tokenizer = tokenizer

        # initialize an empty trainer state (if resuming from a checkpoint,
        # will be updated later)
        self.state = TrainerState()

        if self.args.log_on_main_rank_only:
            # TODO this doesn't seem to work?
            logger.addFilter(logging_utils.FilterMainRankOnly())

    def train(self, resume_from_checkpoint: bool|Path = False) -> None:
        """
        Training loop of the trainer.

        Args:
            resume_from_checkpoint (bool or Path): Whether training should
                resume from a checkpoint or start from scratch. If yes, then
                training resumes from the checkpoint with the latest
                modification time saved in `checkpoints_dir` (see
                `TrainingArguments`).

        """
        # number of examples in dataloader (globally, not per GPU); an example
        # is a sequence of length context_len
        num_examples = len(self.train_dataloader) * self.args.per_device_batch_size

        # weight update steps per epoch (globally, not per GPU)
        num_update_steps_per_epoch = len(self.train_dataloader) // self.args.gradient_accumulation_steps

        # total number of weight update steps
        max_steps = self.args.num_train_epochs * num_update_steps_per_epoch

        # global batch size
        global_batch_size = self.args.per_device_batch_size * self.args.gradient_accumulation_steps * dist_utils.get_world_size()

        fp16_dtype = torch.bfloat16 if torch.cuda.is_bf16_supported() else torch.float16

        loss_fn = torch.nn.CrossEntropyLoss()

        self.progress = tqdm(range(max_steps), disable=(not self.args.show_progress) or dist_utils.get_rank() != 0)

        # scaler used for mixed precision fp16 training on GPU
        # TODO, with bf16, you don't need a scaler
        self.scaler = torch.cuda.amp.GradScaler(enabled=self.args.use_fp16)

        # initialize a Trainer state (will be overriden later if resuming from
        # checkpoint)
        self.state = TrainerState()

        # potentially resume from saved checkpoint
        if isinstance(resume_from_checkpoint, bool) and resume_from_checkpoint:
            last_checkpoint: Path | None = checkpoint_utils.get_last_checkpoint_dir(self.args.checkpoints_dir)
            if last_checkpoint is not None:
                logger.info(f"Resuming training from checkpoint {str(last_checkpoint)}")

            # TODO check what model.state_dict, opt.state_dict etc. really is,
            # and see how to replicate for trainer state
            # TODO possibly move trainer state and trainer args to a separate
            # trainer utils, since state should most likely be held separate
            # from behaviour (Trainer) anyway, and also because TrainerState and
            # checkpoint_utils have a circular dependency
            self.state = TrainerState.from_json_file(os.path.join('checkpoint-1500', 'trainer_state.json'))

            # load model from .pt
            self.model.load_state_dict(torch.load(os.path.join('checkpoint-1500', 'model.pt')))

            #trainer_state.to_json_file(os.path.join('checkpoint-1500', 'trainer_state.json'))
            # TODO write this function and check it
            # TODO also write cu.save_checkpoint()
            # checkpoint_utils.load_from_checkpoint(
            #     last_checkpoint,
            #     self.state,
            #     self.model,
            #     self.optimizer,
            #     self.scheduler,
            #     #self.scaler, # ?
            # )

        self.model.train()

        logger.info("***** Running training *****")
        logger.info(f"  Total number of examples in train dataloader = {num_examples:,}")
        logger.info(f"  Num epochs = {self.args.num_train_epochs:,}")
        logger.info(f"  Instantaneous batch size per device = {self.args.per_device_batch_size:,}")
        logger.info(f"  Gradient Accumulation steps = {self.args.gradient_accumulation_steps}")
        logger.info(f"  Global batch size (w. distributed & accumulation) = {global_batch_size:,}")
        logger.info(f"  Total optimization steps = {max_steps:,}")

        # start training for num_train_epochs (or less if resumed from checkpoint)
        for epoch in range(self.state.current_epoch, self.args.num_train_epochs):

            # needed for distributed sampler RNG state
            if hasattr(self.train_dataloader, "set_epoch"):
                self.train_dataloader.set_epoch(epoch)

            # if resume from checkpoint, update learning rate
            for _ in range(0, self.state.global_steps):
                self.args.lr_scheduler.step()

            for step, inputs in enumerate(self.train_dataloader):

                # if resume from checkpoint, skip the first batches of data in
                # the train dataloader
                if step < (self.state.global_steps % (self.state.current_epoch + 1)) * self.args.gradient_accumulation_steps:
                    continue

                inputs = inputs['input_ids']
                inputs = inputs.to(self.args.device)

                with torch.cuda.amp.autocast(dtype=fp16_dtype,
                                             enabled=self.args.use_fp16):
                    logits = self.model(inputs)

                    labels = inputs[..., 1:].contiguous().view(-1)
                    logits = logits[..., :-1, :].contiguous().view(
                        -1,
                        self.model.module.vocab_size
                        if dist_utils.is_dist_available_and_initialized() # TODO this should probably be a hasattr(self.model, 'module')
                        else self.model.vocab_size
                    )

                    loss = loss_fn(logits, labels)

                    tr_loss = loss.item()
                    loss = loss / self.args.gradient_accumulation_steps # normalize to account for gradient accumulation

                # TODO only for fp16, not fp32 or bf16
                self.scaler.scale(loss).backward()

                # update only after gradient_accumulation_steps
                if (step + 1) % self.args.gradient_accumulation_steps == 0:
                    # Note: This will ignore the last few batches of the dataset
                    # when grad_acc_steps > 1 and the number of batches isn't
                    # divisible by grad_acc_steps

                    self.state.global_steps += 1

                    # gradient clipping
                    self.scaler.unscale_(self.args.optimizer)
                    nn.utils.clip_grad_norm_(self.model.parameters(),
                                             max_norm=self.args.max_grad_norm)

                    self.scaler.step(self.args.optimizer)
                    self.scaler.update()
                    self.args.optimizer.zero_grad()

                    self.args.lr_scheduler.step()
                    # lr = self.args.lr_scheduler.get_last_lr()[0]

                    if self.state.global_steps % self.args.log_steps == 0:
                        logger.info(f"Loss is {tr_loss:,}")

                    print(f'step: {self.state.global_steps}, save: {self.args.save_steps}')
                    if self.state.global_steps % self.args.save_steps == 0:
                        print(f'rank: {dist_utils.get_rank()}')
                        if dist_utils.get_rank() == 0: # TODO this only works for DDP, not FSDP
                            # save checkpoint
                            print(f'saving checkpoint at step {self.state.global_steps}')
                            os.makedirs(os.path.join(self.args.checkpoints_dir, f'checkpoint-{self.state.global_steps}'))
                            self.state.to_json_file(os.path.join(self.args.checkpoints_dir, os.path.join(f'checkpoint-{self.state.global_steps}', 'trainer_state.json')))
                            torch.save(self.model.state_dict(),
                                       os.path.join(self.args.checkpoints_dir,
                                       os.path.join(f'checkpoint-{self.state.global_steps}', 'model.pt')))

                    self.progress.update(1)

    def save_model(self, save_dir: Path) -> None:
        """
        Save model and tokenizer to a directory.

        Args:
            save_dir (Path): Path to save directory.

        """
        pass

    def save_logs(self, log_dir: Path) -> None:
        """
        Save training logs to a directory.

        Args:
            log_dir (Path): Path to log directory.

        """
        pass
