import os
import logging
import contextlib

import torch.distributed as dist


logger = logging.getLogger(__name__)


def init_process_group():
    if 'RANK' not in os.environ or \
        'WORLD_SIZE' not in os.environ or \
        'MASTER_ADDR' not in os.environ or \
        'MASTER_PORT' not in os.environ:
        # at this point, we're pretty sure we are not in a distributed
        # environment
        pass
    else:
        dist.init_process_group('nccl')


def destroy_process_group():
    if 'RANK' not in os.environ or \
        'WORLD_SIZE' not in os.environ or \
        'MASTER_ADDR' not in os.environ or \
        'MASTER_PORT' not in os.environ:
        # at this point, we're pretty sure we are not in a distributed
        # environment
        pass
    else:
        dist.destroy_process_group()


def is_dist_available_and_initialized():
    if not dist.is_available():
        return False

    if not dist.is_initialized():
        return False

    return True


def get_rank():
    if not is_dist_available_and_initialized():
        return 0

    return dist.get_rank()


def get_local_rank():
    try:
        # this should be automatically detected when running with torchrun
        local_rank = int(os.environ['LOCAL_RANK'])
    except:
        # if you run by hand, make sure to correctly set ONLY ONE OF:
        # - LOCAL_RANK=<x>
        #   OR
        # - CUDA_VISIBLE_DEVICES=<x>
        # , where x is one of the GPU's available on the system
        local_rank = 0
    return local_rank


def get_world_size():
    if not is_dist_available_and_initialized():
        return 1

    return dist.get_world_size()


def is_main_rank():
    return get_rank() == 0


@contextlib.contextmanager
def main_process_first():
    rank = get_rank()
    if rank == 0:
        try:
            logger.info(f"Rank 0 start work")
            yield
        finally:
            logger.info(f"Rank 0 done with work")
            if is_dist_available_and_initialized():
                dist.barrier(device_ids=[get_local_rank()])
    else:
        try:
            logger.info(f"Rank {rank} waiting for rank 0 to process work")
            if is_dist_available_and_initialized():
                dist.barrier(device_ids=[get_local_rank()])
        finally:
            logger.info(f"Rank {rank} done waiting for rank 0")
            yield
