import logging

from torch.nn.parallel import DistributedDataParallel as DDP
from transformers import AutoTokenizer
from datasets import load_from_disk

from optimus.models.optimus import OptimusConfig, OptimusTransformer
from optimus.utils import dist_utils


logger = logging.getLogger(__name__)


def load_and_chunk_dataset(data_dir, seq_len):
    tokenized_datasets = load_from_disk(data_dir)

    # split dataset into chunks of seq_len
    def split_text_fn(examples):
        # concatenate texts into batch
        concatenated_examples = {
            input_ids: [token for sublist in examples[input_ids] for token in sublist]
            for input_ids in examples.keys()
        }
        total_length = len(concatenated_examples[next(iter(examples.keys()))])

        # drop last chunk
        total_length = (total_length // seq_len) * seq_len

        # split in chunks of size seq_len
        result = {
            input_ids: [text[i:i + seq_len] for i in range(0, total_length, seq_len)]
            for input_ids, text in concatenated_examples.items()
        }
        return result

    # apply text splitting into batches
    tokenized_datasets = tokenized_datasets.map(
        split_text_fn,
        batched=True,
        num_proc=6,
        remove_columns=tokenized_datasets['train'].column_names
    )
    tokenized_datasets = tokenized_datasets.select_columns(['input_ids'])

    logger.info("Result: ")
    logger.info(tokenized_datasets)

    ctx_len = len(tokenized_datasets['train'][0]['input_ids'])
    n_batches = len(tokenized_datasets['train'])
    logger.info(f"Dataset info:")
    logger.info(f"  - context length: {ctx_len}")
    logger.info(f"  - number of batches (train set): {n_batches}")
    logger.info(f"  - total number of tokens: {ctx_len * n_batches}")

    return tokenized_datasets


def create_tokenizer(tokenizer_name):
    tokenizer = AutoTokenizer.from_pretrained(
        f"./local_tokenizers/{tokenizer_name}",
        local_files_only=True
    )
    if tokenizer_name == 'gpt2':
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.unk_token = tokenizer.eos_token
        tokenizer.bos_token = tokenizer.eos_token

    return tokenizer


def create_model(config_file, device):
    config = OptimusConfig.from_json_file(config_file)
    model = OptimusTransformer(config)
    model.to(device)

    if dist_utils.is_dist_available_and_initialized():
        model = DDP(
            model,
            device_ids=[dist_utils.get_local_rank()],
            output_device=dist_utils.get_local_rank(),
            gradient_as_bucket_view=True,
        )

    logger.info(model)

    _total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    logger.info(f"Number of trainable model parameters: {_total_params}")

    return model
